/* cplans - fr 2005
   simple critical path plan scheduler

   cplans version 0.3  Copyright (C) 2009 Frank Reglin

   cplans is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   cplans is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with cplans; see the file COPYING.
*/

#ifdef LINUX
#define _fullpath(result,input,len) realpath(input,result)
#else
#include <io.h>
#endif

#include <math.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#include "util.h"
#include "date.h"




#define VERSION "0.4 (fr) 2005-2018"
/*
#   input/output syntax:
#   ====================
#   DATE = YYYY-MM-DD
#       2000 <= YYYY <= 2099
#
#   # comment     - does not appear in output
#
#   note comment  - does appear in output
#
#   project start DATE [end DATE]
#       one and only one project statement is needed, it gives the
#       global start and end dates.
#
#   res RESID use USEFACTOR
#       USEFACTOR = effort/duration
#
#   task TASKID prio PRIO res RESID effort EFFORT [fixed] [start DATE] [end DATE]
#       PRIO : high number = high priority
#       fixed = use start/end dates
#
#   dep TASKID1 TASKID2
#       means: task 1 must end before task 2 starts
#
#  Each resource can only work on one task at a time.
#  One permanent "background" task is modeled by means of the use factor.
#  If this is not sufficient you can split the real resources into
#  several logical resources with appropriate use factors.
#
#  All calculations are in days.
#  Holidays are not considered, yet.
#  Weekends are, somehow: by the factor (7/5) in the duration calculation.
#
#  If you want to calculate in durations directly, use resources with use
#  factor 1.4 to compensate the weekend calculation.
#
#  Vacations can be entered as fixed tasks with effort 0 but
#  start date != end date.
#
#  Milestones can be entered as tasks with effort 0 and
#  start date = end date (only one of the dates must be given).
#  In this case a resource is not needed.
#
*/

#define MAXTASK 1000
#define MAXRES   100
#define MAXDEP  5000
#define MAXIDLEN  63
#define MAXLINES ((MAXTASK+MAXRES+MAXDEP)*3)
#define MAXNOTECHARS (MAXLINES*200)
#define MAXPRIO 60000   /* must be < 2**16 - 2000 */
#define MAXEMPTYLINES 1 /* in output */
#define WHITESPACE " \t\n\r"
#define MAXSTRINGLEN 4000
#define TEMPCOLLENTRIES 4

/* message types: */
#define M_FATAL   0 /* program cannot continue                                    */
#define M_ERROR   1 /* error in input file                                        */
#define M_WARNING 2 /* something strange in input file, might be an error         */
#define M_HINT    3 /* something inportant to note                                */
#define M_INFO1   4 /* some more or less useful information the program found out */
#define M_INFO2   5 /* less useful than M_INFO1                                   */
/* FATAL and EEROR result in exit code != 0,
   FATAL, ERROR and WARNING are printed to stderr */

/* task flags: */
#define TASK_FIXED    1  /* task at fixed dates */
#define TASK_START    2  /* start date given    */
#define TASK_END      4  /* end date given      */
#define TASK_SOURCE   8  /* has successor(s)    */
#define TASK_DRAIN   16  /* has predecessor(s)  */
#define TASK_CIRC    32  /* used in circdeptask */

/* lines kinds: */
#define EMPTY 0
#define PRJ   1
#define RES   2
#define TSK   3
#define DEP   4
#define NOT   5

#define STRING(str)  #str


typedef struct
{
   char  id[MAXIDLEN+1];
   double usable;    /* part of work time that can be used for this plan */
   short * calendar; /* resource calendar: for each project day:
                         0 = free, >0 = task index (one and only one task),
                        <0 = collsision = -index into collision table  */
} RESOURCE;

typedef struct
{
   char id[MAXIDLEN+1];
   unsigned short flags;
   unsigned short  prio;
   int  effort;  // man days
   int  res;     // 0 or index into resource[]
   int  duration;// days
   int  start;   // date
   int  end;     // date
   int  eps;     // earliest possible start, can be before given start forfixed tasks
   int  epe;     // earliest possible end
   int  lps;     // latest possible start
   int  lpe;     // latest possible end
   int  numpred; // number of predecessors
   int  numsucc; // number of successors
   short predecessor[MAXTASK];
   short successor[MAXTASK];
} TASK;

typedef struct
{
   int src;  // index into task[]
   int drn;  // index into task[]
} DEPENDENCY;

typedef struct
{
   int kind;
   int idx; // index into resource[] / task[] / dependency[] / notes[]
} LINE;

typedef struct
{
   char * stmtname;
   char * (*function)( void );
} HANDLER;

typedef struct _tempcoll
{
   struct _tempcoll * next;
   short tsk[TEMPCOLLENTRIES];
} TEMPCOLL; // temporary collision entry for max TEMPCOLLENTRIES tasks

typedef struct
{
   short task;
   short prio;
   short permstart;// permutation: start index
   short permnum;  // permutation: # of entries in collision[]
} COLLISION;

#ifdef CPLANS_GLOBALS
#define extern /* */
#endif

extern char * inputfn;     // input file name
extern int  errors;        // number of errors
extern int  maxdate;       // last date the date fucntions can handle, 2099
extern int  infolevel;     // info verbosity
extern int  maxresidlen;   // max resource name length
extern int  maxprio;       // max prio value used
extern char ignorefixed;   // flag. iff != 0 only vacation tasks are left fixed
extern char ignoreprio;    // flag. iff != 0 all prios will be set to 0
extern char ignorerescoll; // flag. iff != 0 stop after pass 1
extern char take1st;       // flag. iff != 0 stop after 1st permutation

extern TASK prj; /* overall project (one and only one) */

extern RESOURCE resource[MAXRES+1]; // [0] not used
extern int maxres; // max index used

extern TASK task[MAXTASK+1]; // [0] not used
extern int  maxtask; // max index used

extern DEPENDENCY dependency[2*MAXDEP+1]; // [0] not used, 2* to have some room for internal adding...
extern int maxdep; // max index used
extern int newdep; // dependencies added to resolve collisions


/* notes are just strings stored one after the other: */
extern char notes[MAXNOTECHARS];
extern int  nextnote; // index where to put next

extern LINE line[MAXLINES+1]; // [0] not used
extern int  maxlin; // max index used

extern COLLISION * collision;// the global collision table

extern TEMPCOLL ** tempcoll; // (dynamic) array of linear lists of TEMPCOLLs
extern int  collentries;     // counts # of collision entries (tasks and trailing zeros)
/*
  More details about collision tables:
  If one resource has to do 2 or more tasks at the same time
  this is called a collision.
  Collisions can happen on each day of the project.
  Note that a task never collides on it's first day. Think
  of tasks always beginning and ending in the evening.
  Each resource has a project calendar that contains an
  integer number for each day. This is 0 (nothing to do),
  a task number (index into task[]) or a collision
  (negated index into collision[]).

  Temporary collision lists: For each day and each resource
  we create a linear list of all colliding tasks. The
  index is just the index into tempcoll[] where the root of
  the list is stored. The index can be calculated directly
  from date and resource.
  resource[].calendar[]-->tempcoll[].tsk[]-->...

  The final collision table has several parts:
  The .task array uses 3 or more entries for each
  collision, one for each colliding task (just the task
  number) and a 0 to mark the end of this collision.
  The index of the first entry of a collision in the array
  represents that collision in the calendar.
  (In the beginning we do not know how many collisions we
  have and how long the individual lists must be.
  Therefore we build temporary lists first and create
  the final array when all collisions are known and
  in the temporary lists.)

  The .priority entries correspondent directly to the .task
  entries, they just contain the task priority. Fixed tasks
  is given a high priority here.

  The .perm... are special: These entries do not correspondent
  to the other elements of the same index. They are in the
  same array just because the number of elements is about
  the same, at least not bigger.
  In these entries we store all permutation groups:
  If there are tasks of the same priority inside of one
  collision they are called permutation group.
  During the scheduling attempts each permutaion group is
  being scheduled independently.
  Each permutation group is represented by 2 values:
  .permstart is the index into collision[] where the group begins,
  .permnum is the number of group members.
*/


/* proto types */
int  taskindex( char * taskid );
int  resindex( char * resid );
int  circdeptask( TASK * task, int t );
int  resourcecalendars( int pass, TASK * prj, TASK * task, int * t1, int * t2 );
void printresourcecalendars( char * prefix, TASK * prj );
void sortcollbyprio( int a, int z );


void readinput( FILE * inf );
void plan( void );

