/* cplans - fr 2005
   simple critical path plan scheduler
   objhdl.c - object handling functions

   cplans version 0.3  Copyright (C) 2009 Frank Reglin

   cplans is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   cplans is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with cplans; see the file COPYING.
*/

#include "cplans.h"


int taskindex( char * taskid )
/* ------------------------------------------------------ */
{
int i;

   for( i = 1; i <= maxtask; i++ )
   {
      if( strcmp(taskid,task[i].id) == 0 )
         return i;
   }
   return 0; /* not found */
}


int resindex( char * resid )
/* ------------------------------------------------------ */
{
int i;

   for( i = 1; i <= maxres; i++ )
   {
      if( strcmp(resid,resource[i].id) == 0 )
         return i;
   }
   return 0; /* not found */
}


int circdeptask( TASK * task, int t )
/* ------------------------------------------------------ */
/* returns 0 iff task t or one of his predecessors has 
   no circular dependency. Otherwise returen the task id
   of a task that has circular dependencies.           */
{
int i, tp, tc;

   if( (task[t].flags & TASK_DRAIN) == 0 )
      return 0;  /* task t has no predecessors */

   if( task[t].flags & TASK_CIRC )
      return t; 

   task[t].flags |= TASK_CIRC;

   for( i = 0; i < task[t].numpred; i++ )
   {
      tp = task[t].predecessor[i];
      tc = circdeptask( task, tp );
      if( tc )
         return tc;
   }

   task[t].flags &= ~TASK_CIRC;
   return 0;
}


static int newcollision( int relday, int res, int t1, int t2 )
/* ------------------------------------------------------ */
/* starts a new temporary collision table,
   returns index into tempcoll[].          */
{
int idx;

   if( !tempcoll )
   {
      tempcoll = alloc( (prj.duration+1)*(maxres+1)*sizeof(TEMPCOLL *) );
      collentries = 0;
   }
   idx = res * (prj.duration+1) + relday;
   tempcoll[idx] = alloc( sizeof(TEMPCOLL) );
   tempcoll[idx]->tsk[0] = t1;
   tempcoll[idx]->tsk[1] = t2;
   collentries += 3; // 2 tasks + 0 at the end
   return idx;
}


static void appendtask( int c, int t )
/* ------------------------------------------------------ */
/* appends task t to the temporary collision c
   (c is index into tempcoll[])                */
{
TEMPCOLL * tc;
int i;

   tc = tempcoll[c];
   while( tc->next )
      tc = tc-> next; // now we are in the last existing list element
   for( i = 1; i < TEMPCOLLENTRIES; i++ )
   {
      if( tc->tsk[i] == 0 )
         goto freefound;
   }
   // this entry is full, allocate a new one:
   tc->next = alloc( sizeof(TEMPCOLL) );
   tc = tc->next;
   i = 0;
 freefound:
   tc->tsk[i] = t;
   collentries++;
}


static void freetempcoll( TEMPCOLL * tc )
/* ------------------------------------------------------ */
/* deallocates a TEMPCOLL list (recursive) */
{
   if( !tc )
      return;
   freetempcoll( tc->next );
   free( tc );
}


static int cmpcollbytask( const COLLISION * elem1, const COLLISION * elem2 )
/* ------------------------------------------------------ */
{
   return elem1->task - elem2->task;   // small task number 1st
}


static int cmpcollbyprio( const COLLISION * elem1, const COLLISION * elem2 )
/* ------------------------------------------------------ */
{
   return elem2->prio - elem1->prio;   // big priority 1st
}


static int issubcoll( int i1, int i2 )
/* ------------------------------------------------------ */
/*  compares the collisions at the indices given,
    returns 1 iff i1 is a subset of i2                    */
{
int k1, k2;

   while( collision[i2].task )
   {
      /* look for start of i1 in i2: */
      if( collision[i2].task == collision[i1].task )
      {
         /* now i1 must be completely in i2 */
         k1 = i1 + 1;
         k2 = i2 + 1;
         while( collision[k1].task )
         {
            if( collision[k1].task != collision[k2].task )
               goto nexti2;
            k1++;
            k2++;
         }
         return 1;
      }
    nexti2:
      i2++;
   }
   return 0;
}


static int uniquecoll( int idx )
/* ------------------------------------------------------ */
/*  is the collision at collision[idx] a new, unique, one ?
    if yes, return idx,
    if no, return the original index for this collision.
    As a side effect uniquecoll sorts the task numbers in
    collision.                                           */
{
int i, nt;

   // How many tasks ?
   nt = 0;
   for( i = idx; collision[i].task; i++ )
      nt++;
   
   // sort the new entries:
   qsort( &(collision[idx]), nt, sizeof(COLLISION), cmpcollbytask );

   // look for the new collision in the old part of the table
   for( i = 1; i < idx; i++ )
      if( collision[i-1].task == 0 ) // take only those i which point to the first task of a collision
         if( issubcoll( idx, i ) )
            return i;           // not new, we have it already at i

   return idx; // didn't found it, so it's a new one
}


static int collisions( void )
/* ------------------------------------------------------ */
/* writes all temporary collisions into the global
   collision table and updates the resource calendars
   accordingly. returns number of collisions.        */
{
int uc, r, d, t, i;
int idx, nidx, uix, maxidx;
TEMPCOLL * tc;

   if( !collentries )
      return 0; // no collisions, nothing to do

   // allocate global collision table:
   collision = alloc( (collentries+1) * sizeof(COLLISION) );

   // fill all temporary collisions into the global table
   // make sure to have only unique collisions
   // correct resource calendar entries
   uc = 0;
   idx = 1;
   for( r = 0; r <= maxres; r++ )
   {
      for( d = 0; d <= prj.duration; d++ )
      {
         nidx = idx;
         tc = tempcoll[r * (prj.duration+1) + d];
         while( tc )
         {
            for( i = 0; i < TEMPCOLLENTRIES; i++ )
            {
               t = tc->tsk[i];
               if( t == 0 )
                  goto done;
               collision[nidx++].task = t;
            }
            tc = tc->next;
         }
       done:
         if( nidx > idx )
         {
            collision[nidx++].task = 0; // trailing zero
            uix = uniquecoll( idx );
            resource[r].calendar[d] = -uix;
            if( uix == idx )
            {
               // yes, a new unique collision
               uc++;
               idx = nidx;
            }
         }
      }
   }
   collentries = idx; // really used part

   // free the temporay collision lists
   maxidx = (prj.duration+1)*(maxres+1);
   for( idx = 0; idx < maxidx; idx++ )
      freetempcoll( tempcoll[idx] );
   free( tempcoll );
   tempcoll    = 0;

   // return number of (unique) collisions
   return uc;
}


int resourcecalendars( int pass, TASK * prj, TASK * task, int * t1, int * t2 )
/* ------------------------------------------------------------------------ */
/* builds resource calendars and (in pass 1) collision table
   if pass > 1 and there are still collisions these are not
   put into collision[] but a pair of task is returned via 
   t1 and t2
   returns number of collisions.                          */
{
int r, t, d, n;
int ncol;
   
   if( pass > 1 )
   {
      for( r = 0; r <= maxres; r++ )
         if( resource[r].calendar )
            free( resource[r].calendar );
   }

   ncol = *t1 = *t2 = 0;
   for( r = 0; r <= maxres; r++ )
   {
      resource[r].calendar = alloc( (prj->duration + 1) * sizeof(short) ); // day 0 .. day duration
   }
   for( t = 1; t <= maxtask; t++ )
   {
      r = task[t].res;
      for( d = task[t].start + 1; d <= task[t].end; d++ ) // tasks do not collide on their 1st day
      {
         n = resource[r].calendar[d-prj->start];
         if( n == 0 )
         {
            resource[r].calendar[d-prj->start] = t;
         }
         else if( t > n ) /* um die Kollisionen nicht doppelt drin zu haben */
         {
            if( pass > 1 )
            {
               resource[r].calendar[d-prj->start] = -1;
               if( (n != *t1) || (t != *t2) )
               {
                  ncol++;
                  *t1 = n;
                  *t2 = t;
               }
            }
            else
            {
               if( n > 0 ) /* neue Kollision beginnen */
                  resource[r].calendar[d-prj->start] = -newcollision( d-prj->start, r, n, t );
               else
                  appendtask( -n, t );  /* Kollision ergaenzen */
            }
         }
      }
   }
   return pass == 1 ? collisions() : ncol;
}


void printresourcecalendars( char * prefix, TASK * prj )
/* ------------------------------------------------------ */
{
int r, d, n;

   for( r = 1; r <= maxres; r++ )
   {
      printf( "# %s: %-*s ", prefix, maxresidlen, resource[r].id );
      for( d = prj->start; d <= prj->end; d++ )
      {
         n = resource[r].calendar[d-prj->start];
         if( n == 0 )
         {
            printf(".");
         }
         else if( n > 0 )
         {
            printf("%c",task[n].id[0]);
         }
         else
         {
            printf( "@" );
         }
      }
      printf( "\n" );
   }
}


void sortcollbyprio( int a, int z )
/* ------------------------------------------------------ */
/* sort collision[a]..collision[z] by priority */
{
   qsort( &(collision[a]), z-a+1, sizeof(COLLISION), cmpcollbyprio );
}


