/* cplans - fr 2005
   simple critical path plan scheduler
   parser.c - input parser

   cplans version 0.3  Copyright (C) 2009 Frank Reglin

   cplans is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   cplans is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with cplans; see the file COPYING.
*/

#include "cplans.h"


static char * note_stmt( void );
static char * dep_stmt( void );
static char * task_stmt( void );
static char * res_stmt( void );
static char * project_stmt( void );

static HANDLER funtab[] =
{
   { "note"   , note_stmt    },
   { "dep"    , dep_stmt     },
   { "task"   , task_stmt    },
   { "res"    , res_stmt     },
   { "project", project_stmt },
   { 0,         0            }
};

static char * keyword[] =
{
   "note",
   "res", "use",
   "task", "prio", "effort", "fixed", "start", "end",
   "dep",
   "project",
   0
};


static int iskeyword( char * str )
/* ------------------------------------------------------ */
{
int i;

   for( i = 0; keyword[i]; i++ )
      if( strcmp( str, keyword[i] ) == 0 )
         return i+1;

   return 0;
}


static char * project_stmt( void )
/* ------------------------------------------------------ */
/*  project start DATE end DATE
 */
{
char * tok;
char * start;
char * end;

   if( prj.start > 0 )
      return "second project statement, need one and only one";

   start = 0;
   end   = 0;
   while( (tok = strtok( 0, WHITESPACE )) != 0 )
   {
      if( strcmp(tok,"start") == 0 )
      {
         if( start )
            return "'start' given twice";
         start = strtok( 0, WHITESPACE );
         if( !start )
            return "start date missing";
         prj.start = atodate( start );
         if( prj.start == -1 )
            return strf("syntax error in date (%s) -- use YYYY-MM-DD", start);
      }
      else if( strcmp(tok,"end") == 0 )
      {
         if( end )
            return "'end' given twice";
         end = strtok( 0, WHITESPACE );
         if( !end )
            return "end date missing";
         prj.end = atodate( end );
         if( prj.end == -1 )
            return strf("syntax error in date (%s) -- use YYYY-MM-DD", end);
      }
      else
      {
         return strf( "unknown keyword '%s'", tok );
      }
   } /* end of while loop */
   if( prj.end <= 0 )
      prj.end = maxdate;

   if( prj.start <= 0 )
      return "no valid date for start of project";
   if( prj.start >= prj.end )
      return "end date must be after start date";

   line[maxlin].kind = PRJ;
   line[maxlin].idx  = 0;
   return 0;
}


static char * res_stmt( void )
/* ------------------------------------------------------ */
/* res ID use USEFACTOR
 */
{
char * resid;
char * usefact;
char * stop;

   maxres++;
   if( maxres > MAXRES )
      return strf( "too many resources (> %d)", MAXRES);

   resid = strtok( 0, WHITESPACE );
   if( !resid || iskeyword(resid) )
      return "resource ID missing. Syntax: res RESID use USEFACTOR";

   if( strlen(resid) > MAXIDLEN )
      return strf( "resource ID too long (max %d chars)", MAXIDLEN);

   if( strcmp(strtok(0,WHITESPACE),"use") != 0  )
      return "syntax error. Syntax: res RESID use USEFACTOR";

   usefact = strtok( 0, WHITESPACE );
   if( !usefact )
      return "use factor missing";

   resource[maxres].usable = strtod( usefact, &stop );
   if( *stop != 0 )
      return strf("syntax error in floating point number (%s)", usefact);

   if( resource[maxres].usable <= 0.0 )
      return "usefactor must be > 0";

   if( strtok( 0, WHITESPACE ) )
      return "syntax error. Syntax: res RESID use USEFACTOR";

   strcpy( resource[maxres].id, resid );

   line[maxlin].kind = RES;
   line[maxlin].idx  = maxres;
   return 0;
}


static char * task_stmt( void )
/* ------------------------------------------------------ */
/* task ID prio PRIO res RESID effort EFFORT [fixed] [start DATE end DATE]
 */
{
char * tok;
char * taskid;
char * prio;
char * resid;
char * effort;
char * start;
char * end;
char * stop;
int  fixed;
int  prioval;

   maxtask++;
   if( maxtask > MAXTASK )
      return strf( "too many tasks (> %d)", MAXTASK);

   task[maxtask].start = 1; /* 1 Jan 2000 */
   task[maxtask].end   = 1; /* 1 Jan 2000 */
   task[maxtask].flags = 0;

   taskid = strtok( 0, WHITESPACE );
   if( !taskid || iskeyword(taskid) )
      return "task ID missing. Syntax: task ID prio PRIO res RESID effort EFFORT [fixed] [start DATE end DATE]";

   if( strlen(taskid) > MAXIDLEN )
      return strf( "task ID too long (max %d chars)", MAXIDLEN);

   if( taskid[0] == '@' )
      return "task ID must not start with a '@'";

   prio  = 0;
   resid = 0;
   effort= 0;
   fixed = 0;
   start = 0;
   end   = 0;
   while( (tok = strtok( 0, WHITESPACE )) != 0 )
   {
      if( strcmp(tok,"prio") == 0 )
      {
         if( prio )
            return "'prio' given twice";
         prio = strtok( 0, WHITESPACE );
         if( !prio )
            return "priority missing";
         prioval = strtol( prio, &stop, 10 );
         if( *stop != 0 )
            return strf("syntax error in number (%s)", prio);
         if( (prioval < 0) || (prioval > MAXPRIO) )
            return strf("priority must be >= 0 and <= %d",MAXPRIO );
         task[maxtask].prio = prioval;
         maxprio = imax( maxprio, task[maxtask].prio );
      }
      else if( strcmp(tok,"res") == 0 )
      {
         if( resid )
            return "'res' given twice";
         resid = strtok( 0, WHITESPACE );
         if( !resid )
            return "resource ID missing";
         task[maxtask].res = resindex( resid );
         if( !(task[maxtask].res) )
            return strf( "unknown resource ID: %s", resid );
      }
      else if( strcmp(tok,"effort") == 0 )
      {
         if( effort )
            return "'effort' given twice";
         effort = strtok( 0, WHITESPACE );
         if( !effort )
            return "effort missing";
         task[maxtask].effort = strtol( effort, &stop, 10 );
         if( *stop != 0 )
            return strf("syntax error in number (%s)", effort);
      }
      else if( strcmp(tok,"start") == 0 )
      {
         if( start )
            return "'start' given twice";
         start = strtok( 0, WHITESPACE );
         if( !start )
            return "start date missing";
         task[maxtask].start = atodate( start );
         if( task[maxtask].start == -1 )
            return strf("syntax error in date (%s) -- use YYYY-MM-DD", start);
         task[maxtask].flags |= TASK_START;
      }
      else if( strcmp(tok,"end") == 0 )
      {
         if( end )
            return "'end' given twice";
         end = strtok( 0, WHITESPACE );
         if( !end )
            return "end date missing";
         task[maxtask].end = atodate( end );
         if( task[maxtask].end == -1 )
            return strf("syntax error in date (%s) -- use YYYY-MM-DD", end);
         task[maxtask].flags |= TASK_END;
      }
      else if( strcmp(tok,"fixed") == 0 )
      {
         if( fixed++ )
            return "'fixed' given twice";
         task[maxtask].flags |= TASK_FIXED;
      }
      else
      {
         return strf( "unknown keyword '%s'", tok );
      }
   } /* end of while loop */
   if( start && end && (task[maxtask].start > task[maxtask].end) )
      return "end date before start date";

   if( fixed && !start && !end )
      return "no start and/or end date given for fixed task";

   strcpy( task[maxtask].id, taskid );

   line[maxlin].kind = TSK;
   line[maxlin].idx  = maxtask;
   return 0;
}


static char * dep_stmt( void )
/* ------------------------------------------------------ */
/*  dep TASKID1 TASKID2 # means 1 before 2
 */
{
char * taskid;
int tidx[2];
int i;
unsigned short prio1, prio2;

   maxdep++;
   if( maxdep > MAXDEP )
      return strf( "too many dependencies (> %d)", MAXDEP);

   for( i = 0; i < 2; i++ )
   {
      taskid = strtok( 0, WHITESPACE );
      if( !taskid || iskeyword(taskid) )
         return "task ID missing";
      tidx[i] = taskindex( taskid );
      if( !tidx[i] )
         return strf( "unknown task ID: %s", taskid );
   }
   if( strtok( 0, WHITESPACE ) )
      return "syntax error. Syntax: dep TASKID1 TASKID2";

   dependency[maxdep].src = tidx[0];
   dependency[maxdep].drn = tidx[1];

   // Prioritaeten vererben: task1 bekommt mindestens die Prio, die task2 hat:
   prio1 = task[dependency[maxdep].src].prio;
   prio2 = task[dependency[maxdep].drn].prio;
   if( prio1 < prio2 )
      task[dependency[maxdep].src].prio = prio2;

   line[maxlin].kind = DEP;
   line[maxlin].idx  = maxdep;
   return 0;
}


static char * note_stmt( void )
/* ------------------------------------------------------ */
/*  note ....
 */
{
char * note;
int len;

   note = strtok( 0, "\n\r" ); /* everything until end of line */
   len = strlen( note );
   if( (nextnote + len + 1) >= MAXNOTECHARS )
      return "too many notes";

   strcpy( &notes[nextnote], note );
   line[maxlin].kind = NOT;
   line[maxlin].idx  = nextnote;
   nextnote += len + 1;
   return 0;
}


static char * statement( int lineno, char * stmt )
/* ------------------------------------------------------ */
/* returns error string (or NULL, if OK)
 */
{
char * s;
char * err;
HANDLER * ph;
static char buf[MAXSTRINGLEN];

   maxlin = lineno;
   line[lineno].kind = EMPTY;
   if( !stmt || (*stmt == '\n') )  // empty line
      return 0;

   // everything after # is a comment, cut it
   s = strchr( stmt, '#' );
   if( s )
      *s = 0;

   // start tokenizing
   stmt = strtok( stmt, WHITESPACE );
   if( !stmt )
      return 0;    // empty or comment

   // now goto the statement function:
   ph = funtab;
   while( ph->stmtname )
   {
      if( strcmp( stmt, ph->stmtname ) == 0 )
      {
         err = ph->function();
         if( err )
         {
            sprintf( buf, "in %s statement: %s", stmt, err );
            return buf;
         }
         else
            return 0;
      }
      ph++;
   }
   sprintf( buf, "unknown statement: %s", stmt );
   return buf;
}


void readinput( FILE * inf )
/* ------------------------------------------------------ */
{
int lineno;
char * err;
char buf[MAXSTRINGLEN];

   lineno = 0;
   while( fgets( buf, sizeof(buf), inf ) )
   {
      lineno++;
      if( lineno > MAXLINES )
      {
         message( M_ERROR, "too many input lines. max:", STRING(MAXLINES) );
         break;
      }
      err = statement( lineno, buf );
      if( err )
         message( M_ERROR, strf( "%s line %d:", inputfn, lineno ), err );
   }

   fclose( inf );
}


