/* cplans - fr 2005
   simple critical path plan scheduler
   schedule.c - this is the scheduling algorithm

   cplans version 0.3  Copyright (C) 2009 Frank Reglin

   cplans is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   cplans is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with cplans; see the file COPYING.
*/

#include "cplans.h"

static int impossibletask;
static int maxpermgrp;
static short * permute;  // ordered list of integers for the permutaion routines
static short * bestperm; // best permutation
static int bestduration; // duration of the best permutation

static char * errstring[] =
{
         "",
/* -1 */ "(too many predecessors)",
/* -2 */ "(too many successors)",
/* -3 */ "(circular dependency)",
/* -4 */ "(fixed tasks overlap)",
/* -5 */ "(fixed task must start later)",
/* -6 */ "(results in new collision)",
         ""
};


static int calcduration( int t )
/* ------------------------------------------------------ */
/* calculates task duration based on resource usage.
   if 0 and fixed task (= vacation) use end - start (if both are given) */
{
int dur;

   dur = (int)ceil( (task[t].effort / resource[task[t].res].usable) * 7. / 5.);
   if( (dur == 0) && ( (task[t].flags & (TASK_FIXED|TASK_START|TASK_END)) == (TASK_FIXED|TASK_START|TASK_END) ) )
   {
      dur = task[t].end - task[t].start; // vacation
   }
   return dur;
}


static void prepare( void )
/* ------------------------------------------------------ */
/* consistency check + fill empty task elements */
{
int t, r, durcalc, durset;

   /* determine max resource name length: */
   maxresidlen = 0;
   for( r = 0; r <= maxres; r++ )
      maxresidlen = imax( maxresidlen, strlen(resource[r].id) );

   /* check for project start/end */
   if( prj.start <= 0 )
      message( M_FATAL, "project statement missing", "" );

   /* for each task: determine duration: */
   for( t = 1; t <= maxtask; t++ )
   {
      durcalc = calcduration( t );
      durset = durcalc;
      task[t].duration = durset;
      if( task[t].flags & TASK_FIXED )
      {
         if( ignorefixed )
         {
            task[t].flags &= ~TASK_FIXED; /* clear fixed flag */
            message( M_INFO1, "ignoring 'fixed' attribute for task", task[t].id );
         }
         if( (task[t].flags & (TASK_START|TASK_END)) == (TASK_START|TASK_END) )
         {
            durset = task[t].end - task[t].start;
            task[t].duration = durset;
            if( durcalc > durset )
               message( M_WARNING, "resource", strf("%s will be overloaded by factor %.3g in task %s",
                       resource[task[t].res].id, (double)durcalc / (double)durset ,task[t].id) );
            if( durcalc != durset )
               message( M_HINT, "duration of task", strf("%s set to %d days, it should be %d days",
                       task[t].id, durset, durcalc) );
         }
         else if( task[t].flags & TASK_START ) /* start given, end calculated: */
         {
            task[t].end = task[t].start + durset;
            if( task[t].end > maxdate )
               message( M_FATAL, "task ends after 2099-12-31 :", task[t].id );
         }
         else if( task[t].flags & TASK_END )   /* end given, start calculated: */
         {
            task[t].start = task[t].end - durset;
            if( task[t].start < 1 )
               message( M_FATAL, "task starts before 2000-1-1 :", task[t].id );
         }
      }
   }

   /* all fixed tasks should be inside the project boundaries: */
   for( t = 1; t <= maxtask; t++ )
   {
      if( task[t].flags & TASK_FIXED )
      {
         if( task[t].start < prj.start )
         {
            if( task[t].effort > 0 )
               message( M_FATAL, "task starts before project starts :", task[t].id );
            else
            {
               message( M_WARNING, "task starts before project starts, task start adjusted :", task[t].id );
               task[t].start = prj.start;
            }
         }
         if( task[t].end > prj.end )
         {
            message( M_FATAL, "task ends after project ends :", task[t].id );
         }
      }
   }

}


static int epeof( int t, int pass, TASK * prj, TASK * task )
/* ------------------------------------------------------ */
/* calculates eps and epe for task t - recursive
   returns epe
   special handling for fixed tasks */
{
int maxepe, i;
int y1, m1, d1;

   if( task[t].epe == 0 )  /* calculate only if not already done */
   {
      if( (task[t].flags & TASK_DRAIN) == 0 )
      { /* no predecessors, set eps */
         task[t].eps = prj->start;
      }
      else
      {
         /* determine max of epe's of the predecessors: */
         maxepe = 0;
         for( i = 0; i < task[t].numpred; i++ )
            maxepe = imax( maxepe, epeof( task[t].predecessor[i], pass, prj, task ) );
         /* this is eps: */
         task[t].eps = maxepe;
      }
      if( (task[t].flags & TASK_FIXED) && (task[t].start < task[t].eps) )
      {
         if( pass == 1 )
         { /* this task simply cannot be done at the fixed dates given,
              even without looking at resources --> make it a moveable task: */
            date2ymd( task[t].start, &y1, &m1, &d1 );
            message( M_WARNING, task[t].id, strf("cannot start at %d-%02d-%02d -- "
                                            "removing fixed attribute",y1,m1,d1) );
            task[t].duration = calcduration( t );
            task[t].flags &= ~TASK_FIXED; /* this must be after the duration calc */
            task[t].end = task[t].eps + task[t].duration;
            task[t].start = task[t].eps; /* earliest possible date */
         }
         else
         {  /* if this happens at later passes, just store the "impossible" task: */
            if( !impossibletask )
               impossibletask = t;
         }
      }
      /* now eps is ok, epe is easy: */
      task[t].epe = task[t].eps + task[t].duration;

      /* might be even later for fixed tasks: */
      if( (task[t].flags & TASK_FIXED) && (task[t].end > task[t].epe) )
      {
         if( pass == 1 )
            message( M_INFO2, task[t].id, strf("could start %d days earlier (not considering resource assignment)",task[t].end-task[t].epe) );
         task[t].epe = task[t].end;
      }
   }
   return task[t].epe;
}


static int lpsof( int t, TASK * prj, TASK * task )
/* ------------------------------------------------------ */
/* calculates lps and lpe for task t - recursive
   returns lps
   epeof run must be before lpsof run ! */
{
int minlps, i;

   if( task[t].lps == 0 )  /* calculate only if not already done */
   {
      if( (task[t].flags & TASK_SOURCE) == 0 )
      { /* no successors, set lpe */
         task[t].lpe = prj->epe;
      }
      else
      {
         /* determine min of lps's of the successors: */
         minlps = prj->epe;
         for( i = 0; i < task[t].numsucc; i++ )
            minlps = imin( minlps, lpsof( task[t].successor[i], prj, task ) );
         /* this is lpe: */
         task[t].lpe = minlps;
      }
      /* now lpe is ok, lps is easy: */
      task[t].lps = task[t].lpe - task[t].duration;
   }
   return task[t].lps;
}


static int schedule( int pass, TASK * prj, TASK * task, int maxdep )
/* -------------------------------------------------------------- */
/* Critical Path Method

   siehe z.B. Biess, Graphentheorie, Mathematik fuer Ingenieure,
   Naturwissenschaftler, Oekonomen und Landwirte, Band 21/2

   Wir haben hier ein Vorgangsknotennetz.
   Tasks sind die Knoten, Dependencies sind die Boegen.

   prj.start und prj.epe fungieren als Start- bzw. End"task".

   pass = 1 : ohne Beachtung der Resourcen.

   returns project duration or <0 (error)
*/
{
int t, r, d, n;
int ts, td, ns, nd, t1, t2;
int y1s, m1s, d1s, y1e, m1e, d1e, y2s, m2s, d2s, y2e, m2e, d2e;

   /* mark tasks that appear in dependencies, add predecessors and successors */
   for( d = 1; d <= maxdep; d++ )
   {
      ts = dependency[d].src;
      td = dependency[d].drn;
      task[ts].flags |= TASK_SOURCE;
      task[td].flags |= TASK_DRAIN;
      nd = task[td].numpred;
      if( nd >= MAXTASK )
      {
         if( pass == 1 )
            message( M_FATAL, "too many predecessors for task", task[td].id );
         else
            return -1;
      }
      ns = task[ts].numsucc;
      if( ns >= MAXTASK )
      {
         if( pass == 1 )
            message( M_FATAL, "too many successors for task", task[ts].id );
         else
            return -2;
      }
      for( n = 0; n < nd; n++ )
      {
         if( task[td].predecessor[n] == ts )
         {
            if( pass == 1 )
            {
               if( ts && td )
                  message( M_HINT, "no need to define dependency multiple times:", strf("dep %s %s",task[ts].id,task[td].id) );
               /* delete this dependency: */
               dependency[d].src = 0;
               dependency[d].drn = 0;
            }
            goto predok;
         }
      }
      task[td].predecessor[nd] = ts;
      task[td].numpred = nd + 1;
   predok:
      for( n = 0; n < ns; n++ )
      {
         if( task[ts].successor[n] == td )
            goto succok;
      }
      task[ts].successor[ns] = td;
      task[ts].numsucc = ns + 1;
   succok:;
   }

   /* check for circular dependencies: */
   for( td = 1; td <= maxtask; td++ )
   {
      if( (task[td].flags & (TASK_SOURCE|TASK_DRAIN)) == (TASK_SOURCE|TASK_DRAIN) )
      {
         t = circdeptask( task, td );
         if( t )
         {
            if( pass == 1 )
               message( M_FATAL, "circular dependency for task", task[t].id );
            else
               return -3;
         }
      }
   }

   /* check for overlapping of fixed tasks for the same resource (and duration > 0): */
   for( t1 = 1; t1 <= maxtask; t1++ )
   {
      if( (task[t1].flags & TASK_FIXED) && (task[t1].duration != 0) )
      {
         r = task[t1].res;
         /* look for another fixed task for same resource: */
         for( t2 = t1 + 1; t2 <= maxtask; t2++ )
         {
            if( (task[t2].res == r) && (task[t2].flags & TASK_FIXED) && (task[t1].duration != 0) )
            {
               if( ((task[t2].start >= task[t1].start) && (task[t2].start <= task[t1].end))
                || ((task[t1].start >= task[t2].start) && (task[t1].start <= task[t2].end)) )
               {
                  if( pass == 1 )
                  {
                     date2ymd( task[t1].start,&y1s, &m1s, &d1s );
                     date2ymd( task[t1].end,  &y1e, &m1e, &d1e );
                     date2ymd( task[t2].start,&y2s, &m2s, &d2s );
                     date2ymd( task[t2].end,  &y2e, &m2e, &d2e );
                     message( M_FATAL, "fixed tasks overlap:", strf( "%s (%d-%02d-%02d..%d-%02d-%02d) and %s (%d-%02d-%02d..%d-%02d-%02d)",
                            task[t1].id, y1s, m1s, d1s, y1e, m1e, d1e,
                            task[t2].id, y2s, m2s, d2s, y2e, m2e, d2e ) );
                  }
                  else
                     return -4;
               }
            }
         }
      }
   }

/* forward calculation: determine eps and epe */
   impossibletask = 0;
   prj->epe = 0;
   for( t = 1; t <= maxtask; t++ )
      prj->epe = imax( prj->epe, epeof( t, pass, prj, task ) );
   if( impossibletask && (pass > 1) )
      return -5;

   prj->duration = prj->epe - prj->start;
   /* all eps and epe ready, prj_duration is now the minimal
      overall duration, i.e length of the critical path   */

/* backward calculation: determine lps and lpe */
   for( t = 1; t <= maxtask; t++ )
      lpsof( t, prj, task );

/* ok, fill start and end dates: */
   for( t = 1; t <= maxtask; t++ )
   {
      if( (task[t].flags & TASK_FIXED) == 0 )
      {
         task[t].start = task[t].eps;
         task[t].end   = task[t].epe;
      }
   }
   if( prj->end != prj->epe )
   {
      if( pass == 1 )
         message( M_INFO2, "project end date changed", "" );
      prj->end = prj->epe;
   }

   return prj->duration;
}


static int attempt( int try )
/* ------------------------------------------------------ */
/* attempts to schedule according to collision[] and permute[]
   allocates prj and task iff try is set,
   if try == 0 -> use original prj and task arrays
   returns project duration */
{
int i, j, dep, pass;
int t1, t2, ta, tb, prev_ta, prev_tb, ncol;
TASK newprj;
TASK * pprj;
TASK * ptask;
int rc = -7;
static int num = 0;

   /* preparation: */
   if( try )
   {
      pass = 2;
      num++;
      newprj  = prj;
      pprj = &newprj;
      ptask = alloc( (maxtask+1) * sizeof(TASK) );
      memcpy( ptask, task, (maxtask+1) * sizeof(TASK) );
      message( M_INFO2, strf("permutation %3d",num), "" );
   }
   else
   {
      pass = 3;
      pprj = &prj;
      ptask = task;
   }
   for( i = 1; i <= maxtask; i++ )
   {
      ptask[i].numpred = 0;
      ptask[i].numsucc = 0;
   }
   /* define dependencies according to the permutation in permute[] */
   dep = maxdep;
   for( i = 1; i < collentries; i++ )
   {
      if( collision[i-1].task == 0 )
      {
         for( j = i; collision[j+1].task; j++ )
         {
            dep++;
            dependency[dep].src = collision[permute[j]].task;
            dependency[dep].drn = collision[permute[j+1]].task;
            message( M_INFO2, "adding:", strf( "dep %s %s", task[dependency[dep].src].id, task[dependency[dep].drn].id ) );
         }
      }
   }

   /* schedule with these new depedencies */
   prev_ta = prev_tb = 0;
   for( i = 1; i <= maxtask; i++ )
   {
      for( j = 1; j <= maxtask; j++ )
      {
         ptask[j].eps = 0;
         ptask[j].epe = 0;
         ptask[j].lps = 0;
         ptask[j].lpe = 0;
      }
      rc = schedule( pass, pprj, ptask, dep );
      if( rc >= 0 )
      {
         ncol = resourcecalendars( pass, pprj, ptask, &t1, &t2 );
         if( ncol == 0 )
            break; /* end of for loop */
         /* new collision. try to solve by adding one more dependency: */
         // printresourcecalendars( "DEBUG", pprj );;;
         rc = -6;
         /* erstmal sehen, ob eine fixed ist: diese als erste */
         if( task[t1].flags & TASK_FIXED )
         {
            ta = t1; tb = t2;
         }
         else if( task[t2].flags & TASK_FIXED )
         {
            ta = t2; tb = t1;
         }
         else if( (task[t1].eps < ptask[t2].eps) || (task[t1].epe < ptask[t2].epe) )
         {
            ta = t1; tb = t2;
         }
         else
         {
            ta = t2; tb = t1;
         }
         if( (ta == prev_ta) && (tb == prev_tb) )
            break; /* useless to try the same twice */
         prev_ta = ta;
         prev_tb = tb;
         dep++;
         dependency[dep].src = ta;
         dependency[dep].drn = tb;
         if( circdeptask( ptask, ta ) )
         {
            dependency[dep].src = tb;
            dependency[dep].drn = ta;
            if( circdeptask( ptask, ta ) )
               return -6; /* hilft alles nix */
         }
         message( M_INFO2, "adding:", strf( "dep %s %s", task[dependency[dep].src].id, task[dependency[dep].drn].id ) );
      }
   }
   if( try )
   {
      message( rc >= 0 ? M_INFO1 : M_INFO2, strf("result of permutation %d is %d", num, rc), rc >= 0 ? "days" : errstring[-rc] );
      free( ptask );
   }
   return rc;
}


static void permgroup( int ngrp )
/* ------------------------------------------------------ */
/* Loop for one permutation group, calls the next one
   recursively. Why a loop as procedure ?
   We need nested loops over the permutation groups,
   number of groups = nesting level !                */
{
int dur, j;

   do
   {
      if( ngrp == maxpermgrp )
      {
         dur = attempt( 1 );
         if( (dur >= 0) && (dur < bestduration) )
         {
            for( j = 1; j < collentries; j++ )
            {
               bestperm[j] = permute[j];
            }
            bestduration = dur;
            if( take1st )
               return;
         }
      }
      else
      {
         permgroup( ngrp+1 );
         if( take1st && (bestduration != 0x7FFFFFFF) )
            return;
      }
   } while( nextperm( permute + collision[ngrp].permstart, collision[ngrp].permnum ) );
}


void plan( void )
/* ====================================================== */
/* procedure that controls the different scheduling attempts */
{
int t, ncol, nperm, i, j, rc;
int prio, xprio, nprio, fixedprio;
int dum1, dum2;

   /* 1. calculate without looking at resources: */
   prepare();
   schedule( 1, &prj, task, maxdep );

   /* build resource calendars and determine collisions: */
   ncol = resourcecalendars( 1, &prj, task, &dum1, &dum2 );

   if( ncol && !ignorerescoll && ignoreprio )
   {
      for( t = 1; t <= maxtask; t++ )
         task[t].prio = 0;
      message( M_INFO1, "ignoring 'prio' values for all tasks", "" );
   }

   /* put priorities into collision table: */
   fixedprio = (maxprio + 2000) / 1000;
   fixedprio *= 1000;
   for( i = 1; i < collentries; i++ )
   {
      t = collision[i].task;
      if( task[t].flags & TASK_FIXED )
         collision[i].prio = fixedprio;
      else
         collision[i].prio = task[t].prio;
   }
   /* sort collisions by priority and depermine number of
      additional dependencies we will later have to create */
   newdep = 0;
   for( i = 1; i < collentries; i++ )
   {
      if( collision[i-1].task == 0 )
      {
         for( j = i+1; collision[j].task; j++ );
         sortcollbyprio( i, j-1 );
         newdep += j-i-1;
         i = j;
      }
   }

   /* print some info about the collisions: */
   if( infolevel >= 1 )
   {
      printf( "# %s: resource calendars before resource scheduling (C = collision)\n", msgtext[M_INFO2] );
      printresourcecalendars( msgtext[M_INFO1], &prj );
   }
   message( M_INFO1, strf("%d task collisions",ncol), ncol ? "before resource scheduling" : "" );
   if( infolevel >= 2 )
   {
      for( i = 1; i < collentries; i++ )
      {
         t = collision[i].task;
         if( t )
            printf( "# %s:    %-*s  %5d\n", msgtext[M_INFO2], MAXIDLEN, task[t].id, collision[i].prio );
         else
            printf( "# %s:    --\n", msgtext[M_INFO2] );
      }
   }

   if( ncol == 0 )
      return; /* done */

   if( ignorerescoll )
   {
      message( M_INFO1, "no resource scheduling", "- ignoring task collisions" );
      return;
   }

/* Ok, if we are still here we have task collisions, i.e. tasks assigned to
   the same resource at the same time. This is handled by the following
   heuristic way: Each collision (= task group) is ordered by priority, fixed
   tasks have the highest priority. (There can be no more than one fixed task
   in each group, otherwise 'fixed' attribute would have been eliminated in
   pass 1 already.) If there are tasks with the same priority we build all
   sorted permutations. For each of these permutations we try to introduce
   additional dependencies that ensure the task order as given by the sorted
   permutation. And we try to schedule. Of all permutations there should be
   one with the shortest overall project duration, this the one we write out
   as result. If there are more than one with shortest duration, we just
   choose one. If there is none (all permutations result in a schedule error)
   then we report the problem.                                              */

/* the collisions are already sorted, now determine the permutation groups: */
   prio = -2;
   xprio = 0;
   nprio = 0;
   maxpermgrp = 0;
   for( i = 1; i < collentries; i++ )
   {
      if( collision[i].prio == prio ) // gleiche Prio
      {
         nprio++;
      }
      else  // neue Prio
      {
         // hatten wir eine Gruppe ?
         if( nprio >= 2 )
         {
            collision[maxpermgrp].permstart = xprio;
            collision[maxpermgrp].permnum   = nprio;
            maxpermgrp++;
         }
         prio = collision[i].prio;
         xprio = i;
         nprio = 1;
      }
   }

   /* overall number of permutations: */
   nperm = 1;
   for( i = 0; i < maxpermgrp; i++ )
      nperm *= fak( collision[i].permnum );

   message( M_INFO2, strf("will add %d dependencies",newdep), "" );
   if( nperm > 0 )
      message( M_INFO2, strf("in %d permutations",nperm), "" );
   message( M_INFO2, strf("in  %d permutation groups",maxpermgrp), "" );

   if( (maxdep + newdep) > (2*MAXDEP) )
      message( M_FATAL, "not enough space for", strf("adding %d dependencies",newdep) );

   /* prepare the trying of all the permutations: */
   permute  = alloc( (collentries+1) * sizeof(short) );
   bestperm = alloc( (collentries+1) * sizeof(short) );
   bestduration = 0x7FFFFFFF;
   /* begin with the initial sorted variant: */
   for( j = 1; j < collentries; j++ )
   {
      bestperm[j] = permute[j] = j;
   }

   /* now all the permutations: */
   permgroup( 0 );
   if( bestduration == 0x7FFFFFFF )
   {
      message( M_WARNING, "task collisions cannot be resolved", "- no resource scheduling done" );
      newdep = 0;
      return;
   }
   message( M_INFO2, strf("%s result: %d days", take1st ? "" : "best", bestduration), "" );

   /* repeat the best attempt and then we are done */
   for( j = 1; j < collentries; j++ )
   {
      permute[j] = bestperm[j];
   }
   rc = attempt( 0 ); /* repeats best attempt and updates resource calendars */
   if( rc < 0 )
      message( M_FATAL, "internal error 2:", strf("second call of best attempt = %d", rc ) );

   printf( "\n# %s: final resource calendars\n", msgtext[M_INFO1] );
   printresourcecalendars( msgtext[M_INFO1], &prj );
}


